.scv-banner {
    padding: 3px;
    border-radius: 2px;
    font-size: 80%;
    text-align: center;
    color: white;
    background: #d40 linear-gradient(-45deg,
        rgba(255, 255, 255, 0.2) 0%,
        rgba(255, 255, 255, 0.2) 25%,
        transparent 25%,
        transparent 50%,
        rgba(255, 255, 255, 0.2) 50%,
        rgba(255, 255, 255, 0.2) 75%,
        transparent 75%,
        transparent
    );
    background-size: 28px 28px;
}
.scv-banner > a {
    color: white;
}


.scv-sphinx_rtd_theme {
    background-color: #2980B9;
}


.scv-bizstyle {
    background-color: #336699;
}


.scv-classic {
    text-align: center !important;
}


.scv-traditional {
    text-align: center !important;
}
